* enable commenting
$eolcom //

* Read set required input data for sets and parameters for further processing *
// $CALL GDXXRW.EXE Input=Testdaten_small.xlsx Output=data.gdx MaxDupeErrors=150 @Dateneingabe.txt
//$CALL GDXXRW.EXE Input=Testdaten.xlsx Output=data.gdx MaxDupeErrors=150 @Dateneingabe.txt

** Required Sets **
Sets

* Generic Set Definitions
V                set of vertices

* Important Subsets
Depot (V)       set of depots
N (V)           set of customers
M               set of vehicles
;

* Receive important sets
$GDXIN data.gdx
$LOAD V, N, M, Depot
$GDXIN

* Important "artifical"/"indice" sets
alias(V,i, j, ix, ix2, ix3);

SET Arc(V, V)   set of arcs between all vertices V / set.V.set.V /;
SET Edge(V, V)  set of edges between all vertices V / Depot.#V, Customer1.#V, Customer2.#V, Customer3.#V, Customer4.#V/ ;

SET Delta_i_Plus (i,j) set of positive incident arcs / set.V.set.V /;
Delta_i_Plus(i,j)$(sameas(i,j)) = no;
SET Delta_i_Minus (i,j) set of negative incident arcs / set.V.set.V /;
Delta_i_Minus(i,j)$(sameas(i,j)) = no;

SET ss set on set;
SET sss(ss) subset on set ss;

** Required Parameters **
Parameters
c_ij(V,V) costs per arc
q(v)   demand at node i
GZ     big number
;

* Receive important parameters
$GDXIN data.gdx
$LOAD c_ij, q
$GDXIN

GZ = 1000;

* See all input information in a GDX file.
Execute_Unload 'input.gdx', i, j, q, Delta_i_Plus, Delta_i_Minus, N, V, Edge;

Variables
x_ij(V,V)    path in graph taken by vehicle

GOAL
Binary Variable x_ij
;


Equations
objective cost function
constraint_1(V) leave constraint
constraint_2(V) visit constraint
constraint_3 leave depot constraint
;

objective .. GOAL =e= sum((i,j), x_ij(i,j) * c_ij(i,j));
constraint_1(i)$(N(i)) .. sum(j$(Delta_i_Plus(i,j)), x_ij(i,j)) =e= 1;
constraint_2(j)$(N(j)) .. sum(i$(Delta_i_Minus(i,j)), x_ij(i,j)) =e= 1;
constraint_3.. sum(j$(Delta_i_Plus('Depot',j)), x_ij('Depot',j)) =l= card(M);

Model vrp /objective, constraint_1, constraint_2, constraint_3 / ;
Option Limrow=1000;
Option optcr=0;
Solve vrp using mip minimizing GOAL ;

* Export results in results.xslx *
Execute_Unload 'results.gdx', x_ij;
*Execute 'GDXXRW.exe results.gdx O=results.xlsx var=x.L'

set cc /c1*c1000/; alias(cc,ccc); // allow 1000 cuts max

singleton set curcut(cc)  current cut always one element;
set allcuts(cc) total cuts;

set t tours  /t1*t10/;
set alltours(t) active tours;

singleton set tt(t)        contains always one element: the current subtour;

sets tour(i,j,t)  tours
     tourset(i,t)    tour set
     visited(i)   flag whether a city is already visited
     next(j)      contains always one element: the to city
     subtours(t)   subtours
     remainleft(i) flag whether investigation is still needed for follow-up subtour elimination check
     loadviolationtours(t) tours which violate load constraints
     startpoints (j) set of given start points of tours
;
singleton sets
     from(i)      contains always one element: the from city
;

* initialize
from(i)$(ord(i)=1) = yes;    // turn first element on
tt(t)$(ord(t)=1)   = yes;    // turn first element on

* In order to communicate with the MIP solver we need certain conventions
* 1. Cut matrix interface requirement with fixed names
parameters
    rhs_c(cc) right hand side
    sense_c(cc) sense of cut 1=l 2=e 3=g
    numcuts    'number of cuts passed back'

* 2. Nonzeros in cut matrix using the original variable name plus _c
    x_ij_c(cc, i, j);

Equations
    cuts_1(cc) subtour elemination constraint
    cuts_2(cc) subtour elemination constraint
    cuts_3(cc) subtour elemination constraint
         ;

cuts_1(allcuts)$(sense_c(allcuts) = 1) .. sum((i,j), x_ij_c(allcuts,i,j)*x_ij(i,j)) =l= rhs_c(allcuts);
cuts_2(allcuts)$(sense_c(allcuts) = 2) .. sum((i,j), x_ij_c(allcuts,i,j)*x_ij(i,j)) =e= rhs_c(allcuts);
cuts_3(allcuts)$(sense_c(allcuts) = 3) .. sum((i,j), x_ij_c(allcuts,i,j)*x_ij(i,j)) =g= rhs_c(allcuts);

Model vrpcut / objective, constraint_1, constraint_2, constraint_3, cuts_1, cuts_2, cuts_3 /;

// determine subtours
scalar iteration;
scalar startedtours;
scalar rhs_value;
scalar ok boolean value;
scalar loadviolation boolean value for loadviolation;
scalar pendeltour boolean value whether pendeltour given;
scalar load;
Scalar Q_max max load per vehicle;
scalar nextItem;
scalar cutscounter;

Q_max = 400;
cutscounter = 0;

numcuts = 1; // to enter the loop
curcut(cc)$(ord(cc)=1) = yes; // init

while(numcuts > 0 and vrpcut.modelStat <> 10, // exit via break
    
    ok = 1; // init
    iteration = 0; // init
    load = 0; // init
    loadviolation = 0; // init
    pendeltour = 0; // init
    numcuts = 0; //re-init
    
    // initialize
    from(i) = no;
    tt(t)   = no;
    subtours(t) = no;
    alltours(t) = no;
    tt(t)$(ord(t)=1)   = yes;    // turn first element on
    tour(i,j,t)=no;              // set all to no
    tourset(i,t)=no;             // re-init required
    visited(i)=no;               // set all nodes to unvisited
    remainleft(i) = no;         // turn all elements off
    loadviolationtours(t) = no; // turn all elements off
    
    startpoints(j) = no;
    startedtours=0; // init
    
    loop(ix$((not sameas('Depot',ix) and x_ij.l('Depot', ix)>0.5)),
        startedtours= startedtours +1;
        startpoints(ix) = yes;
    );
        
    Display startedtours;
    Display startpoints;

    // tour check
    loop(ix3$(startpoints(ix3) and ok = 1), // only for customers
      tour('Depot',ix3,tt) = yes;   // store sequence in table
      from(ix3) = yes;
      load = sum(from, q(from)); // re-init
      loadviolation = 0; // re-init
    
      loop(i$(N(i) and ok = 1 and not startpoints(i)), // only for customers which are not startpoints
            Display from; // show current from node
            next(j)=no; // clearance
            // find city visited after city 'from'
            loop(j, next(j)$(x_ij.l(from,j)>0.5) = yes); // note: check x.l(from,j)=1 would be dangerous
            Display next;
            tour(from,next,tt) = yes;   // store sequence in table
            visited(from)$(not Depot(from)) = yes;        // mark city 'from' as visited (only if not depot)
            from(j) = next(j);
            if (next('Depot'),
                pendeltour = 1;
                Display pendeltour;
            );
            Display tour;
    
            load = load + sum(next, q(next));
            if (loadviolation <1 and load > Q_max,
                loadviolation = 1;
                loadviolationtours(tt) = yes;
                Display "Tour violated";
            )
            Display load;
            if (sum(next, x_ij.l(next,'Depot')) >0 or pendeltour = 1,
                Display "Tour identified -> close tour and create new instance";
                if (pendeltour <1, 
                    tour(next,'Depot',tt) = yes; // close tour
                    visited(next) = yes; // add to visited list
                );
                pendeltour = 0; //re-init            
      
                if (sum(t$(alltours(t)), 1) <= startedTours, // max number of startedTours
                    alltours(tt) = yes; // save tour
                    tt(t) = tt(t-1); // next tour
                );
                break; // exit loop as tour closed
            else // proceed case
                Display "Proceed with current tour";
            )
        );
        display visited;
    
    );
    
    // preparation for subtour check
    display remainleft;
    remainleft(i)$(N(i) and (not visited(i))) = yes;
    ok = 1; // re-init
    display remainleft;
    nextItem = 1; //reinit

    loop(i$remainleft(i), 
        nextItem = ord(i); // get any first item
        break;
    );
    
    // re-initialize
    from(i) = remainleft(i)$(ord(i)=nextItem); // takes probably last one
    
    // determine remaining potential subtours
    loop(i$(remainleft(i) and ok = 1), // only for customers
    
      Display from; // show current from node
      next(j)=no; // clearance
      // find city visited after city 'from'
      loop(j, next(j)$(x_ij.l(from,j)>0.5) = yes); // note: check x.l(from,j)=1 would be dangerous
      Display next;
      tour(from,next,tt) = yes;   // store sequence in table
      visited(from) = yes;        // mark city 'from' as visited
      from(j) = next(j);
      Display tour;
    
      // subtour: only check for sequences like c1 - c2 - c1 or c1 - c2 - c3 - c1. Depot checks not included.
        if (sum(visited(next)$(not Depot(next)),1)>0, // if already visited. Logic is as follows: if next (not yet marked as visited) is already in visited -> subtour (if not depot)!
            Display "Subtour identified -> enter loop";
            subtours(tt) = yes;
            tt(t) = tt(t-1);  // subtour identified. Swap to new tour ("Singleton Set");
            Display "New tour added", tt;
            Display visited, N;
            if (card(visited) < card(N),
              loop((ix,ix2)$(not Depot(ix) and not visited(ix) and x_ij.l(ix, ix2)>0.5), // find starting point of new subtour
               from(ix) = yes;
               break; // exit loop once finished to improve performance
            ) else
                ok = 0;
            )
            display ok;
        )
    );
    
    display tour;
    display subtours;
    display loadviolationtours;
    numcuts = numcuts + sum(t$subtours(t), 1) + sum(t$loadviolationtours(t), 1); // check for violations
    display numcuts;
    
    if (numcuts = 0,
        // export results in results.xslx * 
        Execute_Unload 'results_cut.gdx', GOAL, x_ij;
        abort.noerror 'Accept solution';
    );
    
   // else: introduce cuts for violatedtours
    // rounded capacity inequalities, cf. Toth, Vigo (2014): Vehicle Routing Problem pp. 48-49.
    loop(t$loadviolationtours(t),
       // cleanage -- remove depots to impose correct cuts
       Display curcut;
       rhs_c(curcut)=-1;
       load = 0; //re-init
       loop((ix,ix2)$(tour(ix,ix2, t)),
            tourset(ix, t)$(not sameas('Depot', ix)) = yes;
       );
       display tourset;
       
       loop(ix$(tourset(ix, t)),
            load = load + q(ix);
       );
       Display load;
       rhs_value = ceil( load/Q_max);
       rhs_c(curcut) = rhs_value;
       sense_c(curcut) = 3; // must be greater than
       Display rhs_c;
       loop(ix$(tourset(ix, t)),
            loop(ix2$(not sameas(ix, ix2) and not tourset(ix2, t)),
                x_ij_c(curcut, ix, ix2) = 1;
            );
       );
       Display x_ij_c, rhs_c, sense_c;
       allcuts(curcut) = yes;   // include this cut in set
       curcut(cc) = curcut(cc-1);
    );
    
$Ontext // -- commented as better cuts possible
    simple version of capacity cuts
    loop(t$loadviolationtours(t),
       // cleanage -- remove depots to impose correct cuts
       Display curcut;
       rhs_c(curcut)=-1;
       rhs_value = sum((ix,ix2)$(tour(ix,ix2,t) and (x_ij.l(ix,ix2) > 0.5) and (not sameas(ix, 'Depot')) and (not sameas(ix2,'Depot'))), 1) - 1;
       rhs_c(curcut) = rhs_value;
       sense_c(curcut) = 1; // must be less
       Display rhs_c;
       loop(tour(i,j,t)$(not sameas(i, 'Depot') and not sameas(j,'Depot')),
          iteration = iteration+ 1;
          x_ij_c(curcut, i, j)$(x_ij.l(i,j) > 0.5) = 1;
       );
       Display iteration;
       Display x_ij_c, rhs_c, sense_c;
       allcuts(curcut) = yes;   // include this cut in set
       curcut(cc) = curcut(cc-1);
    );
$Offtext
    
    // else: introduce cuts for subtours
    loop(t$subtours(t),
       Display curcut;
       rhs_c(curcut)=-1;
       rhs_value = sum((ix,ix2)$(tour(ix,ix2,t) and (x_ij.l(ix,ix2) > 0.5)), 1) - 1;
       rhs_c(curcut) = rhs_value;
       sense_c(curcut) = 1; // must be less
       Display rhs_c;
       loop(tour(i,j,t),
          iteration = iteration+ 1;
          x_ij_c(curcut, i, j)$(x_ij.l(i,j) > 0.5) = 1;
       );
       Display iteration;
       Display x_ij_c, rhs_c, sense_c;
       allcuts(curcut) = yes;   // include this cut in set
       curcut(cc) = curcut(cc-1);
    );
    
    display numcuts, x_ij_c, rhs_c, sense_c;
    // See all input information in a GDX file.
   // Execute_Unload 'nextcuts.gdx', numcuts, allcuts, x_ij_c, rhs_c, sense_c;

    cutscounter = cutscounter + numcuts;
    solve vrpcut using mip minimizing GOAL;
);

